/* ******************************************************************

   author     : puehringer edgar

   compiler   : javac (jdk 1.1.3)

   class      : CalendarEntry
   date       : 1.10.97
   description: a class that provides basic methods to write, read,
                delete and update calendar entries from/to a file

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.io.*;
import java.util.*;

/* ******************************************************************
   top level class declaration
*********************************************************************/

public class CalendarFile {

  // class variable declarations
  public final static int NORMAL         = 0;
  public final static int KEEP_NOTES     = 1;
  public final static int REMOVE_PERSONS = 2;
  public final static long EOF = -1;
  private final static float MAX_FRAGMENTATION = 0.5f;
  private final static String BAK_ENDING = ".bak";

  // instance variable declarations
  private DocVer docVer;
  private File calFile;
  private int nrOfReferences;
  private boolean isClosed;
  private RandomAccessFile randFile;
  private long initialSeekPos;
  private long validEntries;
  private long invalidEntries;

  // constructors
  public CalendarFile(DocVer versionInfo, File file) {
    docVer         = versionInfo;
    calFile        = file;
    init();
  } // end constructor CalendarFile

  // methods
  public synchronized void incRefCounter() {
    nrOfReferences++;
  } // end incRefCounter

  public synchronized boolean isRefered() {
    return (nrOfReferences > 0);
  } // end isRefered

  public synchronized void write(CalendarEntry cEntry)
  throws IOException {

    if (isClosed) open();
    randFile.seek(randFile.length());
    cEntry.outputTo(randFile);
    if (cEntry.isDeleted()) invalidEntries++; else validEntries++;
  } // end write

  public synchronized long read(CalendarEntry cEntry, long seekPos)
  throws IOException {
  // reads an entry from seekPos, returns the new position or EOF

    if (isClosed) {
      if (calFile.exists())
        open();
      else
        return EOF;
      // end if
    } // end if
    if (seekPos < initialSeekPos) seekPos = initialSeekPos;
    randFile.seek(seekPos);
    try {
      cEntry.inputFrom(randFile);
    } catch (EOFException excep) {
      return EOF;
    }
    if (cEntry.isDeleted()) invalidEntries++; else validEntries++;
    return randFile.getFilePointer();
  } // end read

  public synchronized long readAndDelete(CalendarEntry cEntry, long seekPos)
  throws IOException {
  // reads and deletes an object from seekPos, returns the new position or EOF
    long newSeekPos;

    if (isClosed) {
      if (calFile.exists())
        open();
      else
        return EOF;
      // end if
    } // end if
    if (seekPos < initialSeekPos) seekPos = initialSeekPos;
    randFile.seek(seekPos);
    try {
      cEntry.inputFrom(randFile);
    } catch (EOFException excep) {
      return EOF;
    }
    newSeekPos = randFile.getFilePointer();
    randFile.seek(seekPos);
    cEntry.deleteOn(randFile);    
    randFile.seek(newSeekPos);
    invalidEntries++;
    return newSeekPos;
  } // end readAndDelete

  public synchronized boolean delete(CalendarEntry cEntry)
  throws IOException {
  // return true if entry was found and deleted
    return (deleteAndReturn(cEntry) != null);
  } // end delete

  public synchronized CalendarEntry deleteAndReturn(CalendarEntry cEntry)
  throws IOException {
  // return the entry that was found and deleted or null otherwise
    CalendarEntry readedEntry = new CalendarEntry();
    long curPos;

    if (isClosed) {
      if (calFile.exists()) open();
        else return null;
    } // end if
    randFile.seek(initialSeekPos);
    do {
      curPos = randFile.getFilePointer();
      try {
        readedEntry.inputFrom(randFile);
      } catch (EOFException excep) {
        return null;
      } 
      if (readedEntry.isDeleted()) invalidEntries++; else validEntries++;
    } while (!cEntry.equals(readedEntry));
    randFile.seek(curPos);
    cEntry.deleteOn(randFile);
    invalidEntries++;
    return readedEntry;
  } // end deleteAndReturn

  public synchronized CalendarEntry update(CalendarEntry from,
                                           CalendarEntry to,
                                           int Flags) throws IOException {
  CalendarEntry deletedEntry = deleteAndReturn(from);

    if (deletedEntry != null) {
      to = (CalendarEntry)to.clone();
      if ((Flags & KEEP_NOTES) != 0) to.setNotes(deletedEntry.getNotes());
      if ((Flags & REMOVE_PERSONS) != 0) {
        String aStr[] = deletedEntry.getPersons();
        Vector toPersons = new Vector();
        int i;

        for (i=0; i<aStr.length; i++)
          if (to.hasPerson(aStr[i])) toPersons.addElement(aStr[i]);
        aStr = new String[toPersons.size()];
        for (i=0; i<toPersons.size(); i++)
          aStr[i] = (String)toPersons.elementAt(i);
        to.setPersons(aStr);
      } // end if
      write(to);
      from         = deletedEntry;
      deletedEntry = to;
      deletedEntry.setOldVersion(from);
    } // end if
    return deletedEntry;
  } // end update

  public synchronized void decRefCounter()
  throws IOException {
    if (nrOfReferences <= 0) return;
    nrOfReferences--;
    if (nrOfReferences > 0) return;
    if (randFile != null) randFile.close();

    CalendarEntry cEntry = new CalendarEntry();
    float fragmentation = (validEntries + invalidEntries) > 0 ?
                          (float)invalidEntries / (float)(validEntries + invalidEntries) :
                          0;

    if ((fragmentation > MAX_FRAGMENTATION) && (calFile.exists())) {
      // copy file without deleted items
      File fromFile = new File(calFile.getAbsolutePath()+BAK_ENDING);
      if (fromFile.exists()) fromFile.delete();
      File renameFile = new File(calFile.getAbsolutePath());
      renameFile.renameTo(new File(calFile.getAbsolutePath()+BAK_ENDING));
      File toFile = new File(calFile.getAbsolutePath());
      fromFile = new File(calFile.getAbsolutePath()+BAK_ENDING);

      FileInputStream fIn   = new FileInputStream(fromFile);
      DataInputStream in    = new DataInputStream(fIn);
      FileOutputStream fOut = new FileOutputStream(toFile);
      DataOutputStream out  = new DataOutputStream(fOut);
      DocVer dVer;

      try {
        dVer = new DocVer(in);
      } catch (EOFException excep) {}
      docVer.outputTo(out);
      while (true) {
        try {
          cEntry.inputFrom(in);
        } catch (EOFException excep) {
          break;
        }
        if (!cEntry.isDeleted()) cEntry.outputTo(out);
      } // end while
      fIn.close();
      out.flush();
      fOut.close();
    } // end if
    init();
  } // end decRefCounter

  protected void init() {
    nrOfReferences = 0;
    isClosed       = true;
    randFile       = null;
    initialSeekPos = 0;
    validEntries   = 0;
    invalidEntries = 0;
  } // end init

  protected void open() throws IOException {
    boolean exists = calFile.exists();
    DocVer fileVer;

    randFile = new RandomAccessFile(calFile,"rw");
    if (exists) try {
      fileVer = new DocVer(randFile);
      if (!fileVer.equals(docVer)) {
        try { randFile.close(); } catch (IOException e) {}
        randFile = null;
        throw new VersionMismatchException();
      } // end if
    } catch (EOFException excep) {
      // Hm, something is wrong with the file version, overwrite it
      exists = false;
      randFile.seek(0);
    }
    if (!exists) {
      docVer.outputTo(randFile);
    } // end if
    initialSeekPos = randFile.getFilePointer();
    isClosed = false;
  } // end open

  protected void finalize() throws Throwable {
    super.finalize();
    if (nrOfReferences > 0) nrOfReferences = 1;
    decRefCounter();
  } // end finalize

} // end CalendarFile

/* ******************************************************************
   end of file
*********************************************************************/
