/* ******************************************************************

   author     : puehringer edgar

   compiler   : javac (jdk 1.1.3)

   class      : ConnectionThread
   date       : 17.11.97
   description: implements the connection thread for the java
                notes server

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.io.*;
import java.net.*;
import java.util.*;

/* ******************************************************************
   top level class declaration
*********************************************************************/

class ConnectionThread extends Thread {

  // class variable declarations
  private static int numberOfConnections = 0;

  // instance variable declarations
  private Socket s;
  private DataInputStream in;
  private DataOutputStream out;
  private Vector outVector;
  private Vector inVector;
  private PwdTable pwdTable;
  private FilePool filePool;
  private FileDatabase fdb;

  // constructors
  ConnectionThread(Socket clientSocket, PwdTable pwdTable,
                   FilePool filePool) {
    s             = clientSocket;
    this.pwdTable = pwdTable;
    this.filePool = filePool;

    try {
      in  = new DataInputStream(s.getInputStream());
      out = new DataOutputStream(s.getOutputStream());
    } catch (IOException excep) {
      try { s.close(); } catch (IOException e) {}
      System.err.println("Unable to set up socket streams."+excep);
      return;
    }
    outVector = new Vector();
    inVector  = new Vector();
    fdb = new FileDatabase(pwdTable,filePool);
    this.start();
  } // end constructor ConnectionThread;

  // class methods
  public static int getNumberOfConnections() {

    return numberOfConnections;
  } // end getNumberOfConnections

  // methods
  public void run() {
    RequestHeader rh;
    boolean connectionActive = true;

    numberOfConnections++;
    try {
      while (connectionActive) {
        receiveVector();
        outVector.removeAllElements();
        rh = (RequestHeader)inVector.elementAt(0);
        switch (rh.getRequestType()) {
          case RequestHeader.LOGIN:
            fdb.loginCyphered(rh.getUser(),rh.getPasswd());
            outVector.addElement(new ReplyHeader(fdb.getErrorMsg()));
          break;
          case RequestHeader.LOGOUT:
            fdb.logout();
            outVector.addElement(new ReplyHeader(fdb.getErrorMsg()));
          break;
          case RequestHeader.CHANGEPASSWD:
            fdb.changePasswdCyphered(rh.getPasswd());
            outVector.addElement(new ReplyHeader(fdb.getErrorMsg()));
          break;
          case RequestHeader.GETUSERS: {
            String users[] = fdb.getUsers();
            if (fdb.failed())
              outVector.addElement(new ReplyHeader(fdb.getErrorMsg()));
            else
              outVector.addElement(new ReplyHeader(users));
            // end if
          } break;
          case RequestHeader.GETDATES: {
            Enumeration e = fdb.getDates(rh.getUser(),rh.getFrom(),
                                  rh.getTo(),rh.isPeriodical());
            if (fdb.failed() || (e == null)) {
              outVector.addElement(new ReplyHeader(fdb.getErrorMsg()));
            }
            else {
              ReplyHeader reply = new ReplyHeader();
              Streamable sa;

              if (e.hasMoreElements()) reply.setLastInSeq(false);
              outVector.addElement(reply);
              while (e.hasMoreElements()) {
                sa = (Streamable)e.nextElement();
                sa.setLastInSeq(!e.hasMoreElements());
                outVector.addElement(sa);
              } // end while
            } // end if
          } break;
          case RequestHeader.GETMESSAGES:
            Enumeration e = fdb.getMessages();
            if (fdb.failed() || (e == null)) {
              outVector.addElement(new ReplyHeader(fdb.getErrorMsg()));
            }
            else {
              ReplyHeader reply = new ReplyHeader();
              Streamable sa;

              if (e.hasMoreElements()) reply.setLastInSeq(false);
              outVector.addElement(reply);
              while (e.hasMoreElements()) {
                sa = (Streamable)e.nextElement();
                sa.setLastInSeq(!e.hasMoreElements());
                outVector.addElement(sa);
              } // end while
            } // end if
          break;
          case RequestHeader.PUTDATE:
            if (inVector.size() != 2) {
              outVector.addElement(new ReplyHeader("Protocol error"+
                                         "\nIllegal number of arguments"+
                                         " for putDate method"));
            }
            else {
              fdb.putDate((CalendarEntry)inVector.elementAt(1));
              outVector.addElement(new ReplyHeader(fdb.getErrorMsg()));
            } // end if
          break;
          case RequestHeader.DELETEDATE:
            if (inVector.size() != 2) {
              outVector.addElement(new ReplyHeader("Protocol error"+
                                         "\nIllegal number of arguments"+
                                         " for deleteDate method"));
            }
            else {
              fdb.deleteDate((CalendarEntry)inVector.elementAt(1),
                             rh.isForAll());
              outVector.addElement(new ReplyHeader(fdb.getErrorMsg()));
            } // end if
          break;
          case RequestHeader.UPDATEDATE:
            if (inVector.size() != 3) {
              outVector.addElement(new ReplyHeader("Protocol error"+
                                         "\nIllegal number of arguments"+
                                         " for updateDate method"));
            }
            else {
              fdb.updateDate((CalendarEntry)inVector.elementAt(1),
                             (CalendarEntry)inVector.elementAt(2),
                             rh.isForAll());
              outVector.addElement(new ReplyHeader(fdb.getErrorMsg()));
            } // end if
          break;
          case RequestHeader.ENDSESSION:
            fdb.endSession();
            outVector.addElement(new ReplyHeader(fdb.getErrorMsg()));
            connectionActive = false;
          break;
          default:
            outVector.addElement(new ReplyHeader("Protocol error"+
                                       "\nClient has sent a unknown request"));
        } // end switch
        sendVector();
      } // end while
    } catch (IOException excep) {}
    finally { try { s.close(); } catch (IOException excep) {} }
    if (numberOfConnections > 0) numberOfConnections--;
  } // end run

  private void sendVector() throws IOException {
    // writes all elements of the outVector to the out stream
    int n          = outVector.size();
    int i;

    for (i=0; i<n; i++) {
      ((Streamable)outVector.elementAt(i)).outputTo(out);
    } // end for
    out.flush();
  } // end sendVector

  private void receiveVector() throws IOException {
    // reads the request header and attached objects from the in stream
    boolean isFirst = true;

    inVector.removeAllElements();
    do {
      if (isFirst) {
        inVector.addElement(new RequestHeader(in));
        isFirst = false;
      }
      else {
        inVector.addElement(new CalendarEntry(in));
      } // end if
    } while (!((Streamable)inVector.elementAt(inVector.size()-1)).isLastInSeq());
  } // end receiveVector

} // end ConnectionThread

/* ******************************************************************
   end of file
*********************************************************************/
