/* ******************************************************************

   author     : puehringer edgar

   compiler   : javac             

   class      : DateChooserCanvas
   date       : 28.11.97
   description: a canvas that allows you to choose a date
   modified   : 9.2.1998 - new constructors for locale scheme added
                           weekday decodeing changed to international
                           standard

*********************************************************************/

/* ******************************************************************
   import statements
*********************************************************************/

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.text.SimpleDateFormat;

/* ******************************************************************
   top level class declaration
*********************************************************************/

public class DateChooserCanvas extends Canvas implements ItemSelectable {

  // class variable declarations
  private final static int DEF_ASCENT = 10;
  private final static int DEF_CHAR_W = 6;

  // instance variable declarations
  private String[] dayShortcuts;
  private Calendar calendar;
  private Calendar first;
  private Calendar today;
  private Vector listenerVector;
  private Dimension size;
  private boolean sizeAvailable;
  private int ascent;
  private int charWidth;
  private int indexOfDate;
  private Locale locale;
  private int firstDay;
  private SimpleDateFormat formatter;
  private DateChooserCanvas me;        // a synonym for this for use in inner
                                       // class
  // inner classes
  class MouseEventHandler extends MouseAdapter {

    // methods
    public void mouseClicked(MouseEvent e) {
      int w = getSize().width;
      int h = getSize().height;
      int x = e.getX()*8/w;
      int y = e.getY()*7/h;

      if ((x > 0) && (y > 0)) {        // if not in week or dayname area
        int i;
        int oldIndex = indexOfDate;
        ItemEvent ie  = new ItemEvent(me,ItemEvent.ITEM_STATE_CHANGED,
                          calendar,ItemEvent.SELECTED);

        indexOfDate = (y-1)*7+x-1;
        // update visible representation
        highlight(null,w,h,oldIndex);
        highlight(null,w,h,indexOfDate);
        // update calendar
        calendar.add(Calendar.DATE,indexOfDate-oldIndex);
        // notify listeners
        for (i=0; i<listenerVector.size(); i++) {
          ((ItemListener)listenerVector.elementAt(i)).itemStateChanged(ie);
        } // end for
      } // end if
    } // end mouseClicked

  } // end MouseEventHandler

  // constructors
  public DateChooserCanvas() {
    init(new GregorianCalendar(),new GregorianCalendar(),Locale.GERMAN);
  } // end constructor DateChooserCanvas

  public DateChooserCanvas(Calendar calendar) {
    Calendar today = (Calendar)calendar.clone();
    today.setTime(new Date());
    init(calendar,today,Locale.GERMAN);
  } // end constructor DateChooserCanvas

  public DateChooserCanvas(Calendar calendar, Calendar today) {
    init(calendar,today,Locale.GERMAN);
  } // end constructor DateChooserCanvas

  public DateChooserCanvas(Calendar calendar, Locale locale) {
    Calendar today = (Calendar)calendar.clone();
    today.setTime(new Date());
    init(calendar,today,locale);
  } // end constructor DateChooserCanvas

  public DateChooserCanvas(Calendar calendar, Calendar today, Locale locale) {
    init(calendar,today,locale);
  } // end constructor DateChooserCanvas

  // methods
  private void init(Calendar calendar, Calendar today, Locale locale) {
    int i;

    me             = this;
    this.calendar  = (Calendar)calendar.clone();
    this.today     = today;
    this.locale    = locale;
    firstDay       = this.calendar.getFirstDayOfWeek();
    formatter      = new SimpleDateFormat("E",this.locale);
    // initialize day of week array
    dayShortcuts  = new String[7];
    Calendar tmpC = (Calendar)this.calendar.clone();
    while (tmpC.get(Calendar.DAY_OF_WEEK) != firstDay) tmpC.add(Calendar.DATE,-1);
    for (i=0; i<7; i++) {
      dayShortcuts[i] = formatter.format(tmpC.getTime());
      tmpC.add(Calendar.DATE,1);
    } // end for
    formatter      = new SimpleDateFormat("w",this.locale);
    first          = (Calendar)this.calendar.clone();
    listenerVector = new Vector();
    sizeAvailable  = false;
    updateFirst();
    addMouseListener(new MouseEventHandler());
  } // end init

  private void updateFirst() {
    // set first to first of month
    indexOfDate = calendar.get(Calendar.DATE)-1;
    first.setTime(calendar.getTime());
    first.add(Calendar.DATE,-indexOfDate);
    // set first to first day of week before or at first of month
    while (first.get(Calendar.DAY_OF_WEEK) != firstDay) {
      indexOfDate++;
      first.add(Calendar.DATE,-1);
    } // end while
  } // end updateFirst

  public void paint(Graphics g) {
    Font font = new Font(g.getFont().getName(),g.getFont().getStyle(),10);
    Color normalColor = g.getColor();
    Color darkColor   = SystemColor.controlShadow;
    Color lightColor  = SystemColor.controlLtHighlight;
    Calendar c        = (Calendar)first.clone();
    int w             = getSize().width;
    int h             = getSize().height;
    int i,j,x,y;
    int day,month,year;

    today.setTime(new Date());
    day   = today.get(Calendar.DATE);
    month = today.get(Calendar.MONTH);
    year  = today.get(Calendar.YEAR);
    g.setFont(font);
    y = h/14+ascent/2;
    for (i=0; i<7; i++) {
      x = w*5/32+w*i/8;
      g.drawString(dayShortcuts[i],x,y);
    } // end for
    x = w/8-1;
    y = h/7-1;
    g.setColor(darkColor);
    g.drawLine(0,y,w,y);
    g.drawLine(x,y+3,x,h);
    y++;
    g.setColor(lightColor);
    g.drawLine(0,y,w,y);
    g.drawLine(x+1,y+2,x+1,h);
    g.setColor(normalColor);
    for (j=0; j<6; j++) {
      x = w/32;
      y = h*3/14+h*j/7+ascent/2;
      g.setColor(normalColor);
      // draw week of year
      c.add(Calendar.DATE,6);
      if (c.get(Calendar.WEEK_OF_YEAR) == 0) {
        c.add(Calendar.DATE,-6);
        g.drawString(formatter.format(c.getTime()),x,y);
      }
      else {
        g.drawString(formatter.format(c.getTime()),x,y);
        c.add(Calendar.DATE,-6);
      } // end if
      for (i=0; i<7; i++) {
        x = w*5/32+w*i/8;
        if (c.get(Calendar.MONTH) == calendar.get(Calendar.MONTH))
          g.setColor(normalColor);
        else
          g.setColor(darkColor);
        // end if
        g.drawString(Integer.toString(c.get(Calendar.DATE)),x,y);
        if ((c.get(Calendar.DATE) == day) &&
            (c.get(Calendar.MONTH) == month) &&
            (c.get(Calendar.YEAR) == year)) {
          g.setColor(SystemColor.control);
          g.draw3DRect((i+1)*w/8+1,(j+1)*h/7+1,w/8,h/7,false);
          g.setColor(normalColor);
        } // end if
        c.add(Calendar.DATE,1);
      } // end for
    } // end for
    highlight(g,w,h,indexOfDate);
  } // end paint

  private void highlight(Graphics g, int w, int h, int index) {
    if (g == null) g = getGraphics();
    g.setXORMode(Color.white);
    g.fillRect((index%7+1)*w/8+1,(index/7+1)*h/7+1,w/8,h/7);
  } // end highlight

  private Dimension calculateSize() {
    Graphics g = getGraphics();
    Font font;
    FontMetrics fm;

    if (g != null) {
      font = new Font(g.getFont().getName(),g.getFont().getStyle(),10);
      g.setFont(font);
      fm            = g.getFontMetrics();
      ascent        = fm.getAscent();
      charWidth     = fm.charWidth('5');
      sizeAvailable = true;
    }
    else {
      ascent    = DEF_ASCENT;
      charWidth = DEF_CHAR_W;
    } // end if
    return new Dimension(32*charWidth,14*ascent);
  } // end calculateSize

  public Date getDate() {
    return calendar.getTime();
  } // end getDate

  public void setDate(Date date) {
    calendar.setTime(date);
    updateFirst();
    update(getGraphics());
  } // end getDate

  public synchronized void addItemListener(ItemListener l) {
    listenerVector.addElement(l);
  } // end addItemListener

  public synchronized void removeItemListener(ItemListener l) {
    listenerVector.removeElement(l);
  } // end addItemListener

  public Object[] getSelectedObjects() {
    return null;
  } // end getSelectedObject

  public Dimension getPreferredSize() {
    if (!sizeAvailable) size = calculateSize();
    return size;
  } // end getPreferredSize

  public Dimension getMinimumSize() {
    return getPreferredSize();
  } // end getMinimumSize

  public Dimension getMaximumSize() {
    return getPreferredSize();
  } // end getMaximumSize

} // end DateChooserCanvas

/* ******************************************************************
   end of file
*********************************************************************/
